# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import ast
from collections import Counter
from copy import copy
from typing import Any, Dict, Iterable, List, Optional, Tuple
import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal
from mitosheet.ai.ai_utils import (
    fix_up_missing_imports,
    get_code_string_from_last_expression,
    replace_last_instance_in_string,
)

from io import StringIO
from contextlib import redirect_stdout

from mitosheet.errors import MitoError, make_column_exists_error, make_exec_error
from mitosheet.state import DATAFRAME_SOURCE_AI, State
from mitosheet.step_performers.dataframe_steps.dataframe_delete import (
    delete_dataframe_from_state,
)
from mitosheet.types import (
    AITransformFrontendResult,
    ColumnHeader,
    ColumnID,
    DataframeReconData,
    ModifiedDataframeReconData,
)


def is_df_changed(old: pd.DataFrame, new: pd.DataFrame) -> bool:
    try:
        assert_frame_equal(old, new, check_names=False)
        return False
    except AssertionError:
        return True


def exec_for_recon(
    code: str, original_df_map: Dict[str, pd.DataFrame]
) -> DataframeReconData:
    """
    Given some Python code, and a list of previously defined dataframes, this function:
    1. Gets all the newly defined dataframes
    2. Gets all the dataframes that have been modified
    3. Gets the value of the last line of code, if that is a Python expression (e.g. not defining a variable)

    It does all of this while only execing the code a single time, which helps performance.

    To find newly defined dataframes, we save the execution environment state before exec runs,
    and then find any new dataframes after exec runs.

    To find modified dataframes, we:
    1. Detect any dataframe names referenced in the code
    2. Add them to the execution environment
    3. Get their new values after execution
    4. Compare with original values to detect changes

    Finally, getting the value of the last line requires:
    1. Parsing the Python AST
    2. Checking if the last line is an expression
    3. If it is, rebuilding it into a string and saving in a temp variable
    4. Accessing that temp variable after execution
    """
    if len(code) == 0:
        return {
            "created_dataframes": {},
            "deleted_dataframes": [],
            "modified_dataframes": {},
            "last_line_expression_value": None,
            "prints": "",
        }

    # Create deep copies of dataframes to avoid modifying originals
    df_map = {df_name: df.copy(deep=True) for df_name, df in original_df_map.items()}

    # Track initial state to detect new variables later
    locals_before = set(locals().keys())

    # Parse the code to handle the last expression
    try:
        ast_before = ast.parse(code)
    except SyntaxError as e:
        raise make_exec_error(e)

    try:
        last_expression = ast_before.body[-1]
    except:
        raise make_exec_error(Exception("No code was generated"))

    # Check if last line is an expression and modify code if needed
    if not isinstance(last_expression, ast.Expr):
        has_last_line_expression_value = False
    else:
        has_last_line_expression_value = True
        last_expression_string: str = get_code_string_from_last_expression(code)  # type: ignore
        code = replace_last_instance_in_string(
            code, last_expression_string, f"FAKE_VAR_NAME = {last_expression_string}"
        )

    # Find dataframes that might be modified
    potentially_modified_df_names = [
        df_name for df_name in original_df_map if df_name in code
    ]

    # Setup execution environment
    globals_dict = {}
    if "lambda" in code:
        globals_dict["pd"] = pd
        globals_dict["np"] = np

    # Create local variables dictionary for execution
    local_vars = {}
    for df_name in potentially_modified_df_names:
        local_vars[df_name] = df_map[df_name]

    # Execute code and capture output
    output_string_io = StringIO()
    try:
        with redirect_stdout(output_string_io):
            exec(code, globals_dict, local_vars)
    except Exception as e:
        raise make_exec_error(e)

    # Analyze results
    created_dataframes: Dict[str, pd.DataFrame] = {
        name: value
        for name, value in local_vars.items()
        if name not in locals_before
        and name != "locals_before"
        and name != "FAKE_VAR_NAME"
        and isinstance(value, pd.DataFrame)
        and name not in original_df_map
    }

    deleted_dataframes = [
        name
        for name in df_map
        if name not in local_vars and name in potentially_modified_df_names
    ]

    modified_dataframes = {
        name: value
        for name, value in local_vars.items()
        if isinstance(value, pd.DataFrame)
        and name in potentially_modified_df_names
        and is_df_changed(original_df_map[name], value)
    }

    # Get last expression value if it exists
    if has_last_line_expression_value:
        last_line_expression_value = local_vars.get("FAKE_VAR_NAME")
    else:
        last_line_expression_value = None

    return {
        "created_dataframes": created_dataframes,
        "deleted_dataframes": deleted_dataframes,
        "modified_dataframes": modified_dataframes,
        "last_line_expression_value": last_line_expression_value,
        "prints": output_string_io.getvalue(),
    }


PANDAS_HAS_NA = hasattr(pd, "NA")


def is_null_column_header_in_column_headers(
    column_header: ColumnHeader, column_headers: Iterable[ColumnHeader]
) -> bool:

    # First, check for NA. Notably, we need to check that NA is
    # is a attribute of pandas, as it's only available in pandas 1.0+
    if PANDAS_HAS_NA and column_header is pd.NA:
        return any(c is pd.NA for c in column_headers)

    # Then, check for NaT
    if column_header is pd.NaT:
        return any(c is pd.NaT for c in column_headers)

    # Then, check for NaN
    if column_header is np.nan:
        return any(c is np.nan for c in column_headers)

    # Then, check None
    if column_header is None and None in column_headers:
        return True

    return False


def is_possibly_null_column_header_in_column_headers_with_no_nans(
    column_header: ColumnHeader, column_headers_with_no_nans: Iterable[ColumnHeader]
) -> bool:
    """
    Checks if the column header is in the list of column headers, taking special care to handle if the column
    header is null. Notably, assumes that the column_headers have no null values.
    """
    if not pd.isna(column_header):
        return column_header in column_headers_with_no_nans

    return is_null_column_header_in_column_headers(
        column_header, column_headers_with_no_nans
    )


def get_added_column_headers(
    old_column_headers: List[ColumnHeader], new_column_headers: Iterable[ColumnHeader]
) -> List[ColumnHeader]:

    old_non_null = list(filter(lambda ch: not pd.isna(ch), old_column_headers))
    new_non_null = list(filter(lambda ch: not pd.isna(ch), new_column_headers))
    added_non_null = list(filter(lambda ch: ch not in old_non_null, new_non_null))

    new_null = list(filter(lambda ch: pd.isna(ch), new_column_headers))
    old_null = list(filter(lambda ch: pd.isna(ch), old_column_headers))
    added_null = list(
        filter(
            lambda ch: not is_null_column_header_in_column_headers(ch, old_null),
            new_null,
        )
    )

    return added_non_null + added_null


def get_shared_column_headers(
    old_column_headers: List[ColumnHeader], new_column_headers: Iterable[ColumnHeader]
) -> List[ColumnHeader]:

    old_non_null = list(filter(lambda ch: not pd.isna(ch), old_column_headers))
    new_non_null = list(filter(lambda ch: not pd.isna(ch), new_column_headers))
    shared_non_null = list(filter(lambda ch: ch in old_non_null, new_non_null))

    new_null = list(filter(lambda ch: pd.isna(ch), new_column_headers))
    old_null = list(filter(lambda ch: pd.isna(ch), old_column_headers))
    shared_null = list(
        filter(
            lambda ch: is_null_column_header_in_column_headers(ch, old_null), new_null
        )
    )

    return shared_non_null + shared_null


def get_modified_dataframe_recon_data(
    old_df: pd.DataFrame, new_df: pd.DataFrame
) -> ModifiedDataframeReconData:
    """
    Given a dataframe and a modified dataframe, this function tries to figure out what has happened
    to column headers dataframe. Specifically, because our state maps column headers to do others based on column
    id, we need to track which columns are added, which are removed, and which are renamed.
    """

    old_columns = old_df.columns.to_list()
    new_columns = new_df.columns.to_list()

    # If there are duplicated columns in the new dataframe, we throw an error
    # as we don't support this elsewhere in the codebase
    if len(new_columns) != len(set(new_columns)):
        raise MitoError(
            "duplicated_column_headers_error",
            "Duplicated Column Headers",
            "There are duplicated columns in the new dataframe",
            error_modal=False,
        )

    old_df_head = old_df.head(5)
    new_df_head = new_df.head(5)

    rows_added_or_removed = len(old_df) != len(new_df)

    # First, preserving the order, we remove any columns that are in both the old
    # and the new dataframe
    old_columns_without_shared = get_added_column_headers(new_columns, old_columns)
    new_columns_without_shared = get_added_column_headers(old_columns, new_columns)

    # Then, we look through to find any columns that have been simply renamed - simply
    # by comparing to see of column are identical between the two values. We do this
    # just by checking the first 5 values of the dataframe, before doing a direct comparison
    renamed_columns: Dict[ColumnHeader, ColumnHeader] = {}
    for old_ch in old_columns_without_shared:
        old_column = old_df_head[old_ch]
        for new_ch in new_columns_without_shared:
            new_column = new_df_head[new_ch]
            if old_column.equals(new_column) and new_ch not in renamed_columns.values():
                renamed_columns[old_ch] = new_ch

    added_columns = [
        ch
        for ch in new_columns_without_shared
        if not is_possibly_null_column_header_in_column_headers_with_no_nans(
            ch, renamed_columns.values()
        )
    ]
    removed_columns = [
        ch
        for ch in old_columns_without_shared
        if not is_possibly_null_column_header_in_column_headers_with_no_nans(
            ch, renamed_columns
        )
    ]

    shared_columns = get_shared_column_headers(old_columns, new_columns)

    if not rows_added_or_removed:
        modified_columns = [
            ch for ch in shared_columns if not old_df[ch].equals(new_df[ch])
        ]
    else:
        # If rows were added or removed, then we don't want to detect every column as having changed
        # and instead we'd just like to report the row changes. As such, we only compare the rows not added or removed
        # NOTE: this is not perfect, as you may have modified columns and removed rows in one go -- but this
        # is ok for most of what we see
        try:
            if len(old_df) < len(new_df):
                df1 = old_df
                df2 = new_df.loc[old_df.index]
            else:
                df1 = old_df.loc[new_df.index]
                df2 = new_df

            modified_columns = [
                ch for ch in shared_columns if not df1[ch].equals(df2[ch])
            ]
        except IndexError:
            modified_columns = [
                ch for ch in shared_columns if not old_df[ch].equals(new_df[ch])
            ]

    return {
        "column_recon": {
            "created_columns": added_columns,
            "deleted_columns": removed_columns,
            "modified_columns": modified_columns,
            "renamed_columns": renamed_columns,
        },
        "num_added_or_removed_rows": len(new_df) - len(old_df),
    }


def delete_column_id_from_state_metadata(
    state: State,
    sheet_index: int,
    column_id: ColumnID,
) -> State:

    # Update all the state variables removing this column from the state
    del state.column_formulas[sheet_index][column_id]
    # TODO: do we want to remove the formulas
    if column_id in state.df_formats[sheet_index]["columns"]:
        del state.df_formats[sheet_index]["columns"][column_id]

    # Clean up the IDs
    state.column_ids.delete_column_id(sheet_index, column_id)

    return state


def update_state_by_reconing_dataframes(
    state: State,
    sheet_index: int,
    old_df: pd.DataFrame,
    new_df: pd.DataFrame,
    column_headers_to_column_ids: Optional[Dict[ColumnHeader, ColumnID]] = None,
) -> Tuple[State, ModifiedDataframeReconData]:
    """
    This function is the work-horse for modified dataframes. It compares the old dataframe at the index
    to the new dataframe, and then updates the state accordingly -- making sure all the metadata is correct.

    This includes: handling deleted columns, added columns, renamed columns, and modified columns.
    """
    # Check there aren't any duplicated columns in the new dataframe
    c = Counter(new_df.columns)
    most_common = c.most_common(1)
    for ch, count in most_common:
        if count > 1:
            raise make_column_exists_error(ch)

    modified_dataframe_recon = get_modified_dataframe_recon_data(old_df, new_df)

    # Add new columns to the state
    if len(modified_dataframe_recon["column_recon"]["created_columns"]) > 0:
        state.add_columns_to_state(
            sheet_index,
            modified_dataframe_recon["column_recon"]["created_columns"],
            column_headers_to_column_ids=column_headers_to_column_ids,
        )

    # Delete removed columns from the state
    deleted_column_ids = state.column_ids.get_column_ids_by_headers(
        sheet_index, modified_dataframe_recon["column_recon"]["deleted_columns"]
    )
    for column_id in deleted_column_ids:
        delete_column_id_from_state_metadata(state, sheet_index, column_id)

    # Rename renamed columns in the state
    for old_ch, new_ch in modified_dataframe_recon["column_recon"][
        "renamed_columns"
    ].items():
        column_id = state.column_ids.get_column_id_by_header(sheet_index, old_ch)
        state.column_ids.set_column_header(sheet_index, column_id, new_ch)

    # Then, actually set the dataframe
    state.dfs[sheet_index] = new_df

    return state, modified_dataframe_recon


def exec_and_get_new_state_and_result(
    state: State, code: str
) -> Tuple[State, Optional[Any], AITransformFrontendResult]:

    # Fix up the code, so we can ensure that we execute it properly
    code = fix_up_missing_imports(code)

    # Make deep copies of all dataframes here, so we can manipulate them without fear
    # TODO: in the future, we could just do the modified ones
    new_state = state.copy(deep_sheet_indexes=list(range(len(state.dfs))))

    df_map = {df_name: df for df_name, df in zip(new_state.df_names, new_state.dfs)}
    recon_data = exec_for_recon(code, df_map)

    # For all of the added dataframes, we just add them to the state
    for df_name, df in recon_data["created_dataframes"].items():
        new_state.add_df_to_state(df, DATAFRAME_SOURCE_AI, df_name=df_name)

    # For deleted dataframes, we remove them from the state
    for df_name in recon_data["deleted_dataframes"]:
        delete_dataframe_from_state(new_state, new_state.df_names.index(df_name))

    # For modified dataframes, we update all the column variables
    modified_dataframes_recons: Dict[str, ModifiedDataframeReconData] = {}
    for df_name, new_df in recon_data["modified_dataframes"].items():
        sheet_index = new_state.df_names.index(df_name)
        new_state, modified_dataframes_recon = update_state_by_reconing_dataframes(
            new_state, sheet_index, new_state.dfs[sheet_index], new_df
        )
        modified_dataframes_recons[df_name] = modified_dataframes_recon

    # For the last value, if is a dataframe, then add it to the state as well -- unless this dataframe
    # is a _newly_ created dataframe that is already given a name
    last_line_expression_value = recon_data["last_line_expression_value"]
    last_line_expression_code = get_code_string_from_last_expression(code)
    last_line_expression_is_previously_created_dataframe = (
        last_line_expression_code in recon_data["created_dataframes"]
    )
    if (
        isinstance(last_line_expression_value, pd.DataFrame)
        or isinstance(last_line_expression_value, pd.Series)
    ) and not last_line_expression_is_previously_created_dataframe:

        # If we get a series, we turn it into a dataframe for the user
        if isinstance(last_line_expression_value, pd.Series):
            last_line_expression_value = pd.DataFrame(
                last_line_expression_value, index=last_line_expression_value.index
            )

        new_state.add_df_to_state(last_line_expression_value, DATAFRAME_SOURCE_AI)
        # We also need to add this to the list of created dataframes, as we didn't know it's name till now
        recon_data["created_dataframes"][
            new_state.df_names[-1]
        ] = last_line_expression_value

    # If the last line value is a primitive, we return it as a result for the frontend
    result_last_line_value = None
    if (
        isinstance(last_line_expression_value, str)
        or isinstance(last_line_expression_value, bool)
        or isinstance(last_line_expression_value, int)
        or isinstance(last_line_expression_value, float)
        or isinstance(last_line_expression_value, np.number)
    ):
        result_last_line_value = last_line_expression_value

    frontend_result: AITransformFrontendResult = {
        "last_line_value": result_last_line_value,
        "created_dataframe_names": list(recon_data["created_dataframes"].keys()),
        "deleted_dataframe_names": recon_data["deleted_dataframes"],
        "modified_dataframes_recons": modified_dataframes_recons,
        "prints": recon_data["prints"],
    }

    return (new_state, recon_data["last_line_expression_value"], frontend_result)



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from mitosheet.scheduling.github import create_github_pr
from mitosheet.scheduling.schedule_utils import get_automation_files_for_new_automation



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
This module contains functions that upgrade steps from one version
to another version. 

# Why

Imagine a user has a saved analysis that they created with an older
version of mitosheet. The add column step in that saved analysis
takes the parameters (sheet_index, column_name).

Then, we create a new version of mitosheet that changes the add column
step to take the parameters (sheet_index, column_name, column_type).

If we don't upgrade the saved analysis, then when the user loads
the saved analysis, the add column step will be missing the column_type
parameter, and the step will fail to execute. 

Thus, anytime you add a parameter to a step, you must write an upgrade
function that adds the parameter to the step. We do this as we have a 
really strong forwards compatibility guarantee. We _never_ want to break
a user's saved analysis!

# How

1.  Go to the step performer of the step you wish to upgrade. Find the old step version -- 
    e.g. in mitosheet/step_performers/graph_steps/graph.py, the step_version is 4 currently.
    Bump the version number by 1.
2.  Create a new file in this directory called {step}.py, if it doesn't already exist.
3.  In that file, create a function called upgrade_{step}_{old_version_num}_to_{new_version_num}.
4.  In that function, write code that upgrades the step from the old version to the new version.
    In most cases, this will mean adding a new parameter to the step, and setting a default
    value for that parameter.
5.  In the upgrade.py file, edit the STEP_UPGRADES_FUNCTION_MAPPING_NEW_FORMAT dictionary
    to include the new upgrade function you just wrote.
6.  Write a test for the upgrade function you just wrote. See mitosheet/tests/test_upgrade.py --
    where you can add an entry to the UPGRADE_TESTS. (The "version" field isn't too important, 
    you can use whatever other tests are using)
7.  Update the version in test_step_format.py in the tests directory to use the new version. 

When adding a test, you can generate a new saved analysis by running the old version of mitosheet,
and using the old version of the step you are upgrading. Then, open ~/.mito/saved_analyses and find 
the analysis with that ID. 

Then, copy that analysis into mitosheet/tests/saved_analyses - and properly specify what the new 
format should be.
"""


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from mitosheet.mito_flask.v1.flatten_utils import (flatten_mito_backend_to_json, read_backend_state_string_to_mito_backend)
from mitosheet.mito_flask.v1.process_event import process_mito_event


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from typing import Any, Dict, List, Optional, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import get_column_header_list_as_transpiled_code, get_column_header_map_as_code_string, get_column_header_as_transpiled_code
from mitosheet.types import ColumnID


class FillNaCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, sheet_index: int, column_ids: List[ColumnID], fill_method: Dict[str, Any]):
        super().__init__(prev_state)
        self.sheet_index = sheet_index
        self.column_ids = column_ids
        self.fill_method = fill_method
        self.full_dataframe = len(self.prev_state.dfs[self.sheet_index]) == len(self.column_ids)

        self.df_name = self.prev_state.df_names[self.sheet_index]
        self.column_headers = self.prev_state.column_ids.get_column_headers_by_ids(self.sheet_index, self.column_ids)

    def get_display_name(self) -> str:
        return 'Filled NaN Values'
    
    def get_description_comment(self) -> str:
        return f'Filled NaN values in {len(self.column_ids)} columns in {self.df_name}'

    def get_code(self) -> Tuple[List[str], List[str]]:
        fill_method_type = self.fill_method['type']

        # If the user is filling NaN values across the entire dataframe, we generate nice code
        # that does this specifically, rather than just doing the columns they picked

        if fill_method_type == 'value':
            if self.full_dataframe:
                
                return [f"{self.df_name}.fillna({get_column_header_as_transpiled_code(self.fill_method['value'])}, inplace=True)"], []
            else:
                values = {column_header: self.fill_method['value'] for column_header in self.column_headers}
                values_string = get_column_header_map_as_code_string(values) # this function is misnamed, but works for us
                return [f"{self.df_name}.fillna({values_string}, inplace=True)"], []
        else:
            if fill_method_type == 'ffill':
                param_string = "method='ffill'"
            elif fill_method_type == 'bfill':
                param_string = "method='bfill'"
            elif fill_method_type == 'mean':
                if self.full_dataframe:
                    param_string = f"{self.df_name}.mean(numeric_only=False)"
                else:
                    param_string = f"{self.df_name}[columns_to_fill_nan].mean(numeric_only=False)"
            elif fill_method_type == 'median':
                if self.full_dataframe:
                    param_string = f"{self.df_name}.median(numeric_only=False)"
                else:
                    param_string = f"{self.df_name}[columns_to_fill_nan].median(numeric_only=False)"

            if self.full_dataframe:
                return [f"{self.df_name}.fillna({param_string}, inplace=True)"], []
            else:
                column_headers_list_str = get_column_header_list_as_transpiled_code(self.column_headers)
                return [
                    f"columns_to_fill_nan = {column_headers_list_str}",
                    f"{self.df_name}[columns_to_fill_nan] = {self.df_name}[columns_to_fill_nan].fillna({param_string})"
                ], []

        return [], []

    def get_edited_sheet_indexes(self) -> List[int]:
        return [self.sheet_index]


#!/usr/bin/env python
# coding: utf-8
# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE.
"""
The mitosheet package is distributed under the mitosheet
package names on pip.
"""
import json
import os
from glob import glob
from pathlib import Path
from typing import List, Tuple
from setuptools import find_packages, setup

HERE = Path(__file__).parent.resolve()
package_json = json.loads(open('package.json').read())
lab_path = Path(HERE, 'mitosheet', 'labextension')

data_files_spec = [
    # Jupyter extension data files
    ("share/jupyter/labextensions/mitosheet", str(lab_path), "**"),
    ("share/jupyter/labextensions/mitosheet", str(HERE), "install.json"),
]

def get_data_files_from_data_files_spec(
    data_specs: List[Tuple[str, str, str]],
):
    """
    Given tuples of (data_file_path, directory_to_search, pattern_to_find),
    this function will return a list of tuples of (data_file_path, [files])
    in the format that setuptools expects.
    """
    file_data = {}
    for (data_file_path, directory_to_search, pattern_to_find) in data_specs or []:
        # Get the directory to search ready
        if os.path.isabs(directory_to_search):
            directory_to_search = os.path.relpath(directory_to_search)
        directory_to_search = directory_to_search.rstrip("/")
        # Get all non-directory files that match the pattern, searching recursively
        files = [
            f for f in glob(
                Path().joinpath(directory_to_search, pattern_to_find).as_posix(), 
                recursive=True
            ) if not os.path.isdir(f)
        ]
        offset = len(directory_to_search) + 1
        for file_name in files:
            relative_path = str(Path(file_name).parent)[offset:]
            full_data_file_path = Path().joinpath(data_file_path, relative_path).as_posix()
            if full_data_file_path not in file_data:
                file_data[full_data_file_path] = []
            file_data[full_data_file_path].append(file_name)
    # Turn to list and sort by length, to be consistent (and maybe cuz we need to for folder creation?)
    data_files = sorted(file_data.items(), key=lambda x: len(x[0]))
    
    return data_files

data_files = get_data_files_from_data_files_spec(data_files_spec)   


# Read the content of README.md for the long description
with open(HERE / "README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup_args = dict(
    name                    = 'mitosheet',
    version                 = package_json["version"],
    url                     = package_json["homepage"],
    author                  = package_json["author"]["name"],
    author_email            = package_json["author"]["email"],
    description             = package_json["description"],
    license                 = "GNU Affero General Public License v3",
    long_description        = long_description,
    long_description_content_type = "text/markdown",
    packages                 = find_packages(exclude=['deployment']),
    include_package_data     = True,
    package_data             = {'': ['*.js', '*.css', '*.html']},
    data_files               = data_files,
    install_requires=[        
        "jupyterlab~=4.0",
        # We allow users to have many versions of pandas installed. All functionality should
        # work, with the exception of Excel import, which might require additonal dependencies
        'pandas>=0.24.2',
        'analytics-python',
        # Graphing libraries
        'plotly>=4.14.3,<6.0.0',
        'chardet>=3.0.4',
        # For XLSX, reading - we don't fix so works on all python versions
        'openpyxl',
        'packaging'
    ],
    extras_require = {
        'test': [
            'pytest',
            'flake8',
            'types-chardet',
            'types-requests',
            'mypy',
            'pytest_httpserver',
        ],
        'deploy': [
            'wheel==0.42.0', 
            'twine==5.1.1',
            'setuptools==68.0.0'
        ],
        'streamlit': [
            'streamlit>=1.24,<1.32',
        ],
        'optional_feature_dependencies': [
            # According to this documentation (https://github.com/snowflakedb/snowflake-connector-python),
            # snowflake-connect-python requires at least Python 3.7
            'snowflake-connector-python[pandas]; python_version>="3.7"',
            'streamlit>=1.24,<1.32',
            'dash>=2.9',
            "flask"
        ]
    },
    zip_safe                = False,
    python_requires         = ">=3.9",
    platforms               = "Linux, Mac OS X, Windows",
    classifiers             = [
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Framework :: Jupyter",
        "Framework :: Jupyter :: JupyterLab",
        "Framework :: Jupyter :: JupyterLab :: 4",
        "Framework :: Jupyter :: JupyterLab :: Extensions",
        "Framework :: Jupyter :: JupyterLab :: Extensions :: Prebuilt"
    ],
)

if __name__ == '__main__':
    setup(**setup_args)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.


from typing import Any, Dict, List, Optional, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.errors import MitoError
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import (
    get_column_header_list_as_transpiled_code, get_column_header_as_transpiled_code)
from mitosheet.types import ColumnID, ColumnHeader

LOOKUP = 'lookup'
UNIQUE_IN_LEFT = 'unique in left'
UNIQUE_IN_RIGHT = 'unique in right'


class MergeCodeChunk(CodeChunk):


    def __init__(
        self,
        prev_state: State,
        how: str,
        destination_sheet_index: Optional[int],
        sheet_index_one: int, 
        sheet_index_two: int, 
        merge_key_column_ids: List[List[ColumnID]], 
        selected_column_ids_one: List[ColumnID], 
        selected_column_ids_two: List[ColumnID], 
        new_df_name: str,
        optional_code_that_successfully_executed: Optional[Tuple[List[str], List[str]]] = None
    ):
        super().__init__(prev_state)
        self.how: str = how 
        self.destination_sheet_index = destination_sheet_index
        self.sheet_index_one: int = sheet_index_one 
        self.sheet_index_two: int = sheet_index_two 
        self.merge_key_column_ids: List[List[ColumnID]] = merge_key_column_ids 
        self.selected_column_ids_one: List[ColumnID] = selected_column_ids_one 
        self.selected_column_ids_two: List[ColumnID] = selected_column_ids_two

        if optional_code_that_successfully_executed is not None:
            self.optional_code_that_successfully_executed = optional_code_that_successfully_executed

        self.df_one_name = self.prev_state.df_names[self.sheet_index_one]
        self.df_two_name = self.prev_state.df_names[self.sheet_index_two]
        self.new_df_name = new_df_name

    def get_display_name(self) -> str:
        return 'Merged'
    
    def get_description_comment(self) -> str:
        return f'Merged {self.df_one_name} and {self.df_two_name} into {self.new_df_name}'

    def _combine_right_with_merge_code_chunk(self, merge_code_chunk: "MergeCodeChunk") -> Optional["CodeChunk"]:
        """
        We can combine a merge code chunk with the one before it if the destination
        sheet index of the other is the created code index of this step.
        """
        destination_sheet_index = self.destination_sheet_index
        other_destination_sheet_index = merge_code_chunk.destination_sheet_index

        # If both of the merges are overwriting the same destination sheet index, and they are both defined
        if destination_sheet_index is not None and destination_sheet_index == other_destination_sheet_index:
            return MergeCodeChunk(
                self.prev_state,
                merge_code_chunk.how,
                merge_code_chunk.destination_sheet_index,
                merge_code_chunk.sheet_index_one, 
                merge_code_chunk.sheet_index_two, 
                merge_code_chunk.merge_key_column_ids, 
                merge_code_chunk.selected_column_ids_one, 
                merge_code_chunk.selected_column_ids_two, 
                merge_code_chunk.new_df_name,
                merge_code_chunk.optional_code_that_successfully_executed
            )

        # If one of the merges if creating the code chunk that the new one is overwriting, then we can optimize
        # this as well
        created_sheet_index = self.get_created_sheet_indexes()
        if created_sheet_index is not None and created_sheet_index[0] == other_destination_sheet_index:
            return MergeCodeChunk(
                self.prev_state,
                merge_code_chunk.how,
                merge_code_chunk.destination_sheet_index,
                merge_code_chunk.sheet_index_one, 
                merge_code_chunk.sheet_index_two, 
                merge_code_chunk.merge_key_column_ids, 
                merge_code_chunk.selected_column_ids_one, 
                merge_code_chunk.selected_column_ids_two, 
                merge_code_chunk.new_df_name,
                merge_code_chunk.optional_code_that_successfully_executed
            )

        return None

    def combine_right(self, other_code_chunk: "CodeChunk") -> Optional["CodeChunk"]:
        if isinstance(other_code_chunk, MergeCodeChunk):
            return self._combine_right_with_merge_code_chunk(other_code_chunk)
        return None
    
    def combine_left(self, other_code_chunk: "CodeChunk") -> Optional["CodeChunk"]:
        # Because overwriting a merge overwrites all the edits on that merge table
        # we can optimize out any edits that are before the merge 
        # NOTE: if we start carrying edits on merges forward, we should remove this 
        # optimization

        destination_sheet_index = self.destination_sheet_index
        edited_sheet_indexes = other_code_chunk.get_edited_sheet_indexes()

        if edited_sheet_indexes is not None and len(edited_sheet_indexes) == 1 and edited_sheet_indexes[0] == destination_sheet_index:
            return MergeCodeChunk(
                other_code_chunk.prev_state,
                self.how,
                self.destination_sheet_index,
                self.sheet_index_one, 
                self.sheet_index_two, 
                self.merge_key_column_ids, 
                self.selected_column_ids_one, 
                self.selected_column_ids_two, 
                self.new_df_name,
                self.optional_code_that_successfully_executed
            )

        return None

    def get_code(self) -> Tuple[List[str], List[str]]:
        try:
            merge_keys_one: List[ColumnHeader] = self.prev_state.column_ids.get_column_headers_by_ids(self.sheet_index_one, list(map(lambda x: x[0], self.merge_key_column_ids)))
            merge_keys_two: List[ColumnHeader] = self.prev_state.column_ids.get_column_headers_by_ids(self.sheet_index_two, list(map(lambda x: x[1], self.merge_key_column_ids)))
        except KeyError:
            raise MitoError(
                'incompatible_merge_key_error',
                f"Could not find merge key in {self.df_one_name} or {self.df_two_name}",
                to_fix='',
                error_modal=False
            )

        selected_column_headers_one: List[ColumnHeader] = self.prev_state.column_ids.get_column_headers_by_ids(self.sheet_index_one, self.selected_column_ids_one)
        selected_column_headers_two: List[ColumnHeader] = self.prev_state.column_ids.get_column_headers_by_ids(self.sheet_index_two, self.selected_column_ids_two)

        if len(merge_keys_one) == 0 and len(merge_keys_two) == 0:
            return [f'{self.new_df_name} = pd.DataFrame()'], ['import pandas as pd']

        # Now, we build the merge code 
        merge_code = []
        if self.how == 'lookup':
            # If the mege is a lookup, then we add the drop duplicates code
            temp_df_name = 'temp_df'
            merge_code.append(f'{temp_df_name} = {self.df_two_name}.drop_duplicates(subset={get_column_header_list_as_transpiled_code(merge_keys_two)}) # Remove duplicates so lookup merge only returns first match')
            how_to_use = 'left'
        else:
            temp_df_name = self.df_two_name
            how_to_use = self.how

        # If we are only taking some columns, write the code to drop the ones we don't need!
        deleted_columns_one = set(self.prev_state.dfs[self.sheet_index_one].keys()).difference(set(selected_column_headers_one).union(set(merge_keys_one)))
        deleted_columns_two = set(self.prev_state.dfs[self.sheet_index_two].keys()).difference(set(selected_column_headers_two).union(set(merge_keys_two)))

        if len(deleted_columns_one) > 0:
            deleted_transpiled_column_header_one_list = get_column_header_list_as_transpiled_code(deleted_columns_one)
            merge_code.append(
                f'{self.df_one_name}_tmp = {self.df_one_name}.drop({deleted_transpiled_column_header_one_list}, axis=1)'
            )
        if len(deleted_columns_two) > 0:
            deleted_transpiled_column_header_two_list = get_column_header_list_as_transpiled_code(deleted_columns_two)
            merge_code.append(
                f'{self.df_two_name}_tmp = {temp_df_name}.drop({deleted_transpiled_column_header_two_list}, axis=1)'
            )

        # If we drop columns, we merge the new dataframes
        df_one_to_merge = self.df_one_name if len(deleted_columns_one) == 0 else f'{self.df_one_name}_tmp'
        df_two_to_merge = temp_df_name if len(deleted_columns_two) == 0 else f'{self.df_two_name}_tmp'

        # We insist column names are unique in dataframes, so we default the suffixes to be the dataframe names
        suffix_one = self.df_one_name
        suffix_two = self.df_two_name if self.df_two_name != self.df_one_name else f'{self.df_two_name}_2'

        # Finially append the merge
        if self.how == UNIQUE_IN_LEFT:
            merge_code.extend([
                f'{self.new_df_name} = {df_one_to_merge}.merge({df_two_to_merge}[{get_column_header_list_as_transpiled_code(merge_keys_two)}], left_on={get_column_header_list_as_transpiled_code(merge_keys_one)}, right_on={get_column_header_list_as_transpiled_code(merge_keys_two)}, how="left", indicator=True, suffixes=(None, "_y"))',
                f'{self.new_df_name} = {self.new_df_name}[{self.new_df_name}["_merge"] == "left_only"].drop(columns="_merge")[{df_one_to_merge}.columns].reset_index(drop=True)',
            ])
        elif self.how == UNIQUE_IN_RIGHT:
            merge_code.extend([
                f'{self.new_df_name} = {df_two_to_merge}.merge({df_one_to_merge}[{get_column_header_list_as_transpiled_code(merge_keys_one)}], left_on={get_column_header_list_as_transpiled_code(merge_keys_two)}, right_on={get_column_header_list_as_transpiled_code(merge_keys_one)}, how="left", indicator=True, suffixes=(None, "_y"))',
                f'{self.new_df_name} = {self.new_df_name}[{self.new_df_name}["_merge"] == "left_only"].drop(columns="_merge")[{df_two_to_merge}.columns].reset_index(drop=True)',
            ])
        else:      
            merge_code.append(
                f'{self.new_df_name} = {df_one_to_merge}.merge({df_two_to_merge}, left_on={get_column_header_list_as_transpiled_code(merge_keys_one)}, right_on={get_column_header_list_as_transpiled_code(merge_keys_two)}, how=\'{how_to_use}\', suffixes=[\'_{suffix_one}\', \'_{suffix_two}\'])'
            )

        # And then return it
        return merge_code, []

    def get_created_sheet_indexes(self) -> List[int]:
        return [len(self.prev_state.dfs)]
    
    def get_source_sheet_indexes(self) -> List[int]:
        return [self.sheet_index_one, self.sheet_index_two]
    
    def can_be_reordered_with(self, code_chunk: CodeChunk) -> bool:
        """
        A merge either creates or edits a sheet. In either case, we don't
        want to reorder around it as this may cause bugs -- so we don't. 
        """
        return False



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from typing import Any, Dict, List, Optional, Set, Tuple

import pandas as pd

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.promote_row_to_header_code_chunk import \
    PromoteRowToHeaderCodeChunk
from mitosheet.errors import make_invalid_promote_row_to_header
from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.utils import get_param
from mitosheet.types import ColumnHeader
from mitosheet.utils import convert_df_to_parsed_json


def get_should_deduplicate_column_headers(column_headers: List[ColumnHeader]) -> bool:
    """
    Check if the column headers should be deduplicated, taking special care to 
    handle nans.
    """
    return len(set(column_headers)) != len(column_headers) or \
        sum(isinstance(ch, float) and pd.isna(ch) for ch in column_headers) >= 2
        

class PromoteRowToHeaderStepPerformer(StepPerformer):
    """
    Allows you to promote a row to a header (and deletes that row in the process).
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'promote_row_to_header'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:

        post_state, execution_data = cls.execute_through_transpile(
            prev_state, 
            params
        )

        # We make sure that this making of headers will work, and not cause issues 
        # later while trying to convert to json. We throw an error if this causes
        # errors. See this bug: https://github.com/mito-ds/monorepo/issues/267
        try:
            sheet_index: int = get_param(params, 'sheet_index')
            convert_df_to_parsed_json(post_state.dfs[sheet_index].head(1))
        except:
            raise make_invalid_promote_row_to_header()

        return post_state, execution_data

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            PromoteRowToHeaderCodeChunk(
                prev_state, 
                get_param(params, 'sheet_index'), 
                get_param(params, 'index'),
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}
    


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import pandas as pd
import pytest

from mitosheet.scheduling.schedule_utils import (
    get_automation_code, get_automation_files_for_new_automation,
    get_cron_string_from_schedule)
from mitosheet.tests.test_utils import create_mito_wrapper
from mitosheet.tests.decorators import only_on_github_actions


def test_get_automation_code_with_no_params():
    mito = create_mito_wrapper()
    automation_code = get_automation_code(
        mito.mito_backend.steps_manager,
        'automation_name',
        'automation_name',
    )

    assert automation_code == """
def automation_name():
    return 



# Run the automation
automation_name()"""

def ADDONE(x):
    return x + 1 

def test_get_automation_code_with_custom_function_params():

    df = pd.DataFrame({'A': [1, 2, 3]})
    mito = create_mito_wrapper(df, sheet_functions=[ADDONE])
    mito.set_formula('=ADDONE(A1)', 0, 'B', add_column=True)
    automation_code = get_automation_code(
        mito.mito_backend.steps_manager,
        'automation_name',
        'automation_name',
    )

    assert automation_code == """from mitosheet.public.v3 import *
from mitosheet.tests.scheduling.test_schedule_utils import ADDONE

def automation_name(import_dataframe_0):
    import_dataframe_0['B'] = ADDONE(import_dataframe_0['A'].shift(-1, fill_value=0))
    
    return import_dataframe_0



# Run the automation
import_dataframe_0 = r"automation_name/data/df1"
automation_name(import_dataframe_0)"""

def test_get_automation_code(tmp_path):
    mito = create_mito_wrapper()
    df = pd.DataFrame({'A': [1, 2, 3]})

    import_path = str(tmp_path / 'input.csv')
    df.to_csv(import_path, index=False)

    mito.simple_import([import_path])
    mito.add_column(0, 'B', -1)
    output_path = str(tmp_path / 'output.csv')
    output_path_xlsx = str(tmp_path / 'output.xlsx')
    mito.export_to_file('csv', [0], output_path)
    mito.export_to_file('excel', [0], output_path_xlsx)

    automation_code = get_automation_code(
        mito.mito_backend.steps_manager,
        'automation_name',
        'automation_name',
    )
    
    assert automation_code == """import os
import sys
from mitosheet.public.v3 import *
import pandas as pd

def automation_name(file_name_import_csv_0, file_name_export_csv_0, file_name_export_excel_0):
    input = pd.read_csv(file_name_import_csv_0, sep='A')
    
    input['B'] = 0
    
    input.to_csv(file_name_export_csv_0, index=False)
    
    with pd.ExcelWriter(file_name_export_excel_0, engine="openpyxl") as writer:
        input.to_excel(writer, sheet_name="input", index=False)
    
    return input


# Create a folder for this run
export_time = sys.argv[1]
os.makedirs(f"automation_name/runs/{export_time}")

# Run the automation
file_name_import_csv_0 = r"automation_name/data/input.csv"
file_name_export_csv_0 = f"automation_name/runs/{export_time}/file_name_export_csv_0.csv"
file_name_export_excel_0 = f"automation_name/runs/{export_time}/file_name_export_excel_0.xlsx"
automation_name(file_name_import_csv_0, file_name_export_csv_0, file_name_export_excel_0)"""


def test_get_automation_files_returns_correct_files(tmp_path):
    mito = create_mito_wrapper()
    df = pd.DataFrame({'A': [1, 2, 3]})
    import_path = str(tmp_path / 'input.csv')
    df.to_csv(import_path, index=False)
    mito.simple_import([import_path])

    files = get_automation_files_for_new_automation(
        mito.mito_backend.steps_manager,
        'newname',
        'automation_description',
        'automations/newname',
        {
            'type': 'Every Day',
            'time': '09:00'
        }
    )

    assert len(files) == 5
    assert 'automations/newname/automation.py' in files and len(files['automations/newname/automation.py']) > 0
    assert 'automations/newname/requirements.txt' in files and len(files['automations/newname/requirements.txt']) > 0
    assert 'automations/newname/README.md' in files and len(files['automations/newname/README.md']) > 0
    assert 'automations/newname/data/input.csv' in files and len(files['automations/newname/data/input.csv']) > 0
    assert '.github/workflows/newname.yml' in files and len(files['.github/workflows/newname.yml']) > 0


"""
type EveryDayAutomationSchedule = {
    type: 'Every Day',
    time: string
}

type EveryWeekAutomationSchedule = {
    type: 'Every Week',
    dayOfWeek: number,
    time: string
}

type EveryMonthAutomationSchedule = {
    type: 'Every Month',
    dayOfMonth: number,
    time: string
}
"""
CRON_TESTS = [
    (
        {
            'type': 'Every Day',
            'time': '09:00'
        }
        , '0 9 * * *'
    ),
    (
        {
            'type': 'Every Week',
            'dayOfWeek': 1,
            'time': '09:00'
        }
        , '0 9 * * 1'
    ),
    (
        {
            'type': 'Every Month',
            'dayOfMonth': 1,
            'time': '09:00'
        }
        , '0 9 1 * *'
    )
]

@pytest.mark.parametrize('schedule, cron_string', CRON_TESTS)
@only_on_github_actions
def test_cron_string_from_schedule(schedule, cron_string):
    assert get_cron_string_from_schedule(schedule) == cron_string


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

"""
    This code is responsible for making mitosheet the default dataframe display when Jupyter 
    is rendering a dataframe to the output cell. 

    This approach works by overriding the default HTML and plain text formatters for pandas DataFrames in IPython.
    When a DataFrame is displayed, the custom formatter functions are called, which in turn call mitosheet.sheet()
    to render the DataFrame using mitosheet. 
    
    In order for the custom renderers to work, the mitosheet package must first be imported in the user's kernel. 
    To do so, when the Mito extension is activated, we register a kernel status listener that imports the mitosheet package
    whenever the kernel changes.

    This approach still fails the race condition caused when the user clicks "Restart Kernel and run all cells". 
    When doing so, the `import mitosheet` does not get executed until after all of the code cells have been executed. 
    Which means that the mitosheet package is not imported when rendering any dataframes in the notebook. In
    this case, we just default to the pandas dataframe renderer. That is okay.

    _____

    Previous approaches: 
    
    Before landing on this approach, we tried to create a custom dataframe mime renderer. Using the mime renderer
    did not work because: 

    In order to create the mitosheet, we need to execute the mitosheet.sheet() function. To do so, we 
    had to send a new kernel message from the mimerender with the code mitosheet.sheet(df). However, becasue
    the kernel message queue might have had additional messages already queued that edited the df, by the time
    the mitosheet was rendered, it might have displayed a dataframe that reflected future code cell edits instead of the
    current state of the dataframe at the time the code cell with the hanging df was executed. This is not what we want. 

    For a more thorough explanation of attempted solutions, see the comment here: https://github.com/mito-ds/mito/pull/1330#issuecomment-2386428760
    """

def set_dataframe_display_formatters() -> None:

    try: 
        # Since Mito is used in Streamlit which is not a iPython enviornment, 
        # we just wrap this entire thing in a try, except statement 

        from IPython import get_ipython
        import pandas as pd
        import mitosheet

        ip = get_ipython() # type: ignore
        if not ip: # Exit if not in an IPython environment
            return

        # Custom HTML formatter for DataFrames using Mitosheet
        def mitosheet_display_formatter(obj, include=None, exclude=None):
            if isinstance(obj, pd.DataFrame):
                # Render the DataFrame using Mitosheet
                return mitosheet.sheet(obj, input_cell_execution_count=ip.execution_count)
            # Returning None tells Jupyter that this formatter (the text/html one) didn’t produce any output for the given object.
            # This causes Jupyter to “fall back” to lower-priority formatters (like text/plain).
            return None

        # Custom plain text formatter to suppress plain text for DataFrames
        def mitosheet_plain_formatter(obj, p, cycle):
            # TODO: I'm not 100% confident that this is correct, but if I don't overwrite the plain text formatter
            # for dataframes, then I end up getting the mitosheet followed by a printed version of the dataframe.
            if isinstance(obj, pd.DataFrame):
                # Returning None here tells Jupyter not to render anything in the text/plain format 
                # for DataFrames. In this case, Jupyter has already rendered the text/html output, 
                # so returning None for text/plain means “suppress this output,” 
                # preventing Jupyter from rendering the plain text fallback.
                return None
            # For other objects, use the default plain text formatter
            return p.text(obj)

        ip = get_ipython() # type: ignore
        
        html_formatter = ip.display_formatter.formatters['text/html']
        plain_formatter = ip.display_formatter.formatters['text/plain']

        # Register the custom formatters
        html_formatter.for_type(pd.DataFrame, mitosheet_display_formatter)
        plain_formatter.for_type(pd.DataFrame, mitosheet_plain_formatter)
    
    except:
        pass


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

# This folder includes all the code chunks that should _always_ be at the end of the 
# transpiled code. This is useful for things like the formatting, which needs to be
# at the end. Note that these should be transpiled with with the latest state in the
# analysis


from mitosheet.code_chunks.postprocessing.set_dataframe_format_code_chunk import SetDataframeFormatCodeChunk


POSTPROCESSING_CODE_CHUNKS = [
    SetDataframeFormatCodeChunk
]


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from typing import Any, List, Optional, Dict

from openpyxl.styles import Font, PatternFill
from openpyxl.styles import NamedStyle
from openpyxl.formatting.rule import FormulaRule
from openpyxl.worksheet.worksheet import Worksheet
from pandas import DataFrame, ExcelWriter

from mitosheet.excel_utils import get_column_from_column_index
from mitosheet.is_type_utils import is_float_dtype, is_int_dtype
from mitosheet.types import (
    FC_BOOLEAN_IS_FALSE, FC_BOOLEAN_IS_TRUE, FC_DATETIME_EXACTLY,
    FC_DATETIME_GREATER, FC_DATETIME_GREATER_THAN_OR_EQUAL, FC_DATETIME_LESS,
    FC_DATETIME_LESS_THAN_OR_EQUAL, FC_DATETIME_NOT_EXACTLY, FC_EMPTY,
    FC_NOT_EMPTY, FC_NUMBER_EXACTLY,
    FC_NUMBER_GREATER, FC_NUMBER_GREATER_THAN_OR_EQUAL, FC_NUMBER_HIGHEST,
    FC_NUMBER_LESS, FC_NUMBER_LESS_THAN_OR_EQUAL, FC_NUMBER_LOWEST,
    FC_NUMBER_NOT_EXACTLY, FC_STRING_CONTAINS, FC_STRING_DOES_NOT_CONTAIN,
    FC_STRING_ENDS_WITH, FC_STRING_EXACTLY, FC_STRING_NOT_EXACTLY,
    FC_STRING_STARTS_WITH, FC_STRING_CONTAINS_CASE_INSENSITIVE)


# Object to map the conditional formatting operators to the excel formulas
CONDITION_TO_COMPARISON_FORMULA: Dict[str, str] = {
    FC_NUMBER_GREATER: '>',
    FC_NUMBER_LESS: '<',
    FC_NUMBER_NOT_EXACTLY: '<>',
    FC_NUMBER_GREATER_THAN_OR_EQUAL: '>=',
    FC_NUMBER_LESS_THAN_OR_EQUAL: '<=',
    FC_DATETIME_EXACTLY: '=',
    FC_DATETIME_NOT_EXACTLY: '<>',
    FC_DATETIME_GREATER: '>',
    FC_DATETIME_GREATER_THAN_OR_EQUAL: '>=',
    FC_DATETIME_LESS: '<',
    FC_DATETIME_LESS_THAN_OR_EQUAL: '<=',
}

def get_conditional_format_rule(
    filter_condition: str,
    fill: Optional[PatternFill],
    font: Optional[Font],
    filter_value: str,
    cell_range: str,
    column: str
) -> Optional[FormulaRule]:
    # Update the formulas for the string operators
    comparison = CONDITION_TO_COMPARISON_FORMULA.get(filter_condition)
    # If comparing dates, we need to use the DATEVALUE function
    if comparison is not None and 'datetime' in filter_condition:
        formula = [f'{cell_range}{comparison}DATEVALUE("{filter_value}")']
    elif comparison is not None:
        formula = [f'{cell_range}{comparison}{filter_value}']
    elif filter_condition == FC_NUMBER_EXACTLY:
        # Sometimes, users will enter a string for a number filter -- because it is just
        # displayed as an = on the frontend. This is a reasonably thing for them
        # to do -- so we take special care in this case to wrap the filter value in
        # quotes
        try:
            float(filter_value)
            formula = [f'{cell_range}={filter_value}']
        except ValueError:
            formula = [f'{cell_range}="{filter_value}"']
    elif filter_condition == FC_STRING_CONTAINS:
        formula = [f'NOT(ISERROR(FIND("{filter_value}",{cell_range})))']
    elif filter_condition == FC_STRING_CONTAINS_CASE_INSENSITIVE:
        formula = [f'NOT(ISERROR(SEARCH("{filter_value}",{cell_range})))']
    elif filter_condition == FC_STRING_DOES_NOT_CONTAIN:
        formula = [f'ISERROR(SEARCH("{filter_value}",{cell_range}))']
    elif filter_condition == FC_STRING_STARTS_WITH:
        formula = [f'LEFT({cell_range},LEN("{filter_value}"))="{filter_value}"']
    elif filter_condition == FC_STRING_ENDS_WITH:
        formula = [f'RIGHT({cell_range},LEN("{filter_value}"))="{filter_value}"']
    elif filter_condition == FC_BOOLEAN_IS_TRUE:
        formula = [cell_range]
    elif filter_condition == FC_BOOLEAN_IS_FALSE:
        formula = [f'NOT({cell_range})']
    elif filter_condition == FC_STRING_EXACTLY:
        formula = [f'EXACT({cell_range},"{filter_value}")']
    elif filter_condition == FC_STRING_NOT_EXACTLY:
        formula = [f'NOT(EXACT({cell_range},"{filter_value}"))']
    elif filter_condition == FC_EMPTY:
        formula = [f'ISBLANK({cell_range})']
    elif filter_condition == FC_NOT_EMPTY:
        formula = [f'NOT(ISBLANK({cell_range}))']
    elif filter_condition == FC_NUMBER_HIGHEST:
        formula = [f'{column}2>=LARGE({column}:{column},{filter_value})']
    elif filter_condition == FC_NUMBER_LOWEST:
        formula = [f'{column}2<=SMALL({column}:{column},{filter_value})']
    else: return None
    return FormulaRule(fill=fill, font=font, formula=formula)

def add_conditional_formats(
    conditional_formats: Optional[List[Any]],
    sheet: Worksheet,
    df: DataFrame
) -> None:
    if conditional_formats is None:
        return

    for conditional_format in conditional_formats:
        for filter in conditional_format.get('filters', []):
            # Create the conditional formatting color objects
            cond_fill = None
            cond_font = None
            if conditional_format.get('background_color') is not None:
                cond_fill = PatternFill(start_color=conditional_format['background_color'][1:], end_color=conditional_format['background_color'][1:], fill_type='solid')
            if conditional_format.get('font_color') is not None:
                cond_font = Font(color=conditional_format['font_color'][1:])
            if cond_fill is None and cond_font is None:
                continue
            
            # Add the conditional formatting rule to the sheet
            for column_header in conditional_format['columns']:
                column_index = df.columns.tolist().index(column_header)
                column = get_column_from_column_index(column_index)
                cell_range = f'{column}2:{column}{sheet.max_row}'
                column_conditional_rule = get_conditional_format_rule(
                    filter_condition=filter['condition'],
                    fill=cond_fill,
                    font=cond_font,
                    filter_value=f"{filter['value']}",
                    cell_range=cell_range,
                    column=column
                )
                if column_conditional_rule is not None:
                    sheet.conditional_formatting.add(cell_range, column_conditional_rule)


def add_number_formatting(
    number_formats: Optional[Dict[str, str]],
    sheet: Worksheet,
    df: DataFrame
) -> None:
    default_number_format_column_indexes = set(range(len(df.columns)))

    # For all the columns that have user defined number formats 
    # apply those
    if number_formats is not None:
        for column_header, number_format in number_formats.items():
            column_index = df.columns.tolist().index(column_header)
            column = get_column_from_column_index(column_index)
            cell_range = f'{column}2:{column}{sheet.max_row}'
            for cell in sheet[cell_range]:
                cell[0].number_format = number_format

            default_number_format_column_indexes.remove(column_index)

    # For all the float columns that have default number formats
    # and contain numbers, apply the default number format
    # which has comma separators and 2 decimal places
    column_headers = df.columns.tolist()
    for column_index in default_number_format_column_indexes:
        column_header = column_headers[column_index]
        dtype = str(df[column_header].dtype)
        column = get_column_from_column_index(column_index)
        cell_range = f'{column}2:{column}{sheet.max_row}'

        if is_float_dtype(dtype):
            for cell in sheet[cell_range]:
                cell[0].number_format = '#,##0.00'
        if is_int_dtype(dtype):
            for cell in sheet[cell_range]:
                cell[0].number_format = '#,##0'

def add_header_formatting_to_excel_sheet(
    writer: ExcelWriter,
    sheet_name: str,
    header_background_color: Optional[str]=None,
    header_font_color: Optional[str]=None
) -> None:
    """
    Adds formatting to the header row of the sheet_name, based on the formatting the user
    currently has applied in the frontend. 
    """
    workbook = writer.book
    sheet = workbook.get_sheet_by_name(sheet_name)
    
    # Add formatting to the header row   
    header_name = f"{sheet_name}_Header"
    header_format = NamedStyle(name=header_name)
    if header_font_color:
        # Remove the # from the color
        header_format.font = Font(color=header_font_color[1:])
    if header_background_color:
        # Remove the # from the color
        header_format.fill = PatternFill(start_color=header_background_color[1:], end_color=header_background_color[1:], fill_type="solid")
    
    # Only add format if there is a header color or background color
    has_header_formatting = header_background_color or header_font_color 
    if has_header_formatting:
        # Add named styles for the header rows to improve performance
        workbook.add_named_style(header_format)

        # Write the formatting to the sheet
        for col in range(1, sheet.max_column + 1):
            sheet.cell(row=1, column=col).style = header_name

def add_row_formatting_to_excel_sheet(
    writer: ExcelWriter,
    sheet_name: str,
    even_background_color: Optional[str]=None,
    even_font_color: Optional[str]=None,
    odd_background_color: Optional[str]=None,
    odd_font_color: Optional[str]=None
) -> None:
    """
    Adds formatting to the rows of the sheet_name, based on the formatting the user
    currently has applied in the frontend. 
    """
    workbook = writer.book
    sheet = workbook.get_sheet_by_name(sheet_name)

    # Add formatting to the rows
    even_name = f"{sheet_name}_Even"
    odd_name = f"{sheet_name}_Odd"
    even_format = NamedStyle(name=even_name)
    odd_format = NamedStyle(name=odd_name)

    # Remove the # from the colors and define the formatting objects
    if even_background_color:
        even_format.fill = PatternFill(start_color=even_background_color[1:], end_color=even_background_color[1:], fill_type="solid")
    if even_font_color:
        even_format.font = Font(color=even_font_color[1:])
    if odd_background_color:
        odd_format.fill = PatternFill(start_color=odd_background_color[1:], end_color=odd_background_color[1:], fill_type="solid")
    if odd_font_color:
        odd_format.font = Font(color=odd_font_color[1:])
    
    # Only add format if there is a background color or font color
    has_row_formatting = even_background_color or even_font_color or odd_background_color or odd_font_color
    if not has_row_formatting:
        return

    workbook.add_named_style(even_format)
    workbook.add_named_style(odd_format)

    for row in range(2, sheet.max_row + 1):
        for col in range(1, sheet.max_column + 1):
            if row % 2 == 0:
                sheet.cell(row=row, column=col).style = even_name
            else:
                sheet.cell(row=row, column=col).style = odd_name


def add_formatting_to_excel_sheet(
    writer: ExcelWriter,
    sheet_name: str,
    df: DataFrame,
    header_background_color: Optional[str]=None,
    header_font_color: Optional[str]=None,
    even_background_color: Optional[str]=None,
    even_font_color: Optional[str]=None,
    odd_background_color: Optional[str]=None,
    odd_font_color: Optional[str]=None,
    conditional_formats: Optional[list]=None,
    number_formats: Optional[Dict[str, Any]]=None
) -> None:
    """
    Adds formatting to the sheet_name, based on the formatting the user
    currently has applied in the frontend. 

    NOTE: this is a Mito Pro feature.
    """
    workbook = writer.book
    sheet = workbook.get_sheet_by_name(sheet_name)

    add_header_formatting_to_excel_sheet(
        writer=writer,
        sheet_name=sheet_name,
        header_background_color=header_background_color,
        header_font_color=header_font_color
    )

    add_row_formatting_to_excel_sheet(
        writer=writer,
        sheet_name=sheet_name,
        even_background_color=even_background_color,
        even_font_color=even_font_color,
        odd_background_color=odd_background_color,
        odd_font_color=odd_font_color
    )

    add_conditional_formats(conditional_formats, sheet, df)

    add_number_formatting(number_formats, sheet, df)


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import json
import pytest
from mitosheet.mito_backend import MitoBackend, get_mito_backend
import pandas as pd

df = pd.DataFrame({
    'A': [1, 2, 3, 4, 5, 6, 7, 8, 9],
    'B': [1, 2, 3, 4, 5, 6, 7, 8, 9]
})

def test_create_backend_with_column_definitions_works():
    

    column_definitions= [
        [
            {
                'columns': ['A', 'B'],
                'conditional_formats': [{
                    'filters': [{'condition': 'greater', 'value': 5}], 
                    'font_color': '#c30010', 
                    'background_color': '#ffcbd1' 
                }] 
            },
            {
                'columns': ['A'],
                'conditional_formats': [{
                    'filters': [{'condition': 'less_than_or_equal', 'value': 0}], 
                    'font_color': '#FFFFFF', 
                    'background_color': '#000000' 
                }] 
            },
        ]
    ]

    mito_backend = get_mito_backend(df, column_definitions=column_definitions)

    df_formats = mito_backend.steps_manager.curr_step.df_formats
    
    # Still generates valid, empty df_formats objects
    assert df_formats[0]['columns'] == {}
    assert df_formats[0]['headers'] == {}
    assert df_formats[0]['rows']['even'] == {}
    assert df_formats[0]['rows']['odd'] == {}
    assert df_formats[0]['border'] == {}

    # But now has conditional formats
    assert len(df_formats[0]['conditional_formats']) == 2
    assert df_formats[0]['conditional_formats'][0]['columnIDs'] == ['A', 'B']
    assert df_formats[0]['conditional_formats'][0]['filters'] == [{'condition': 'greater', 'value': 5}]
    assert df_formats[0]['conditional_formats'][0]['color'] == '#c30010'
    assert df_formats[0]['conditional_formats'][0]['backgroundColor'] == '#ffcbd1'
    assert df_formats[0]['conditional_formats'][0]['invalidFilterColumnIDs'] == []

    assert df_formats[0]['conditional_formats'][1]['columnIDs'] == ['A']
    assert df_formats[0]['conditional_formats'][1]['filters'] == [{'condition': 'less_than_or_equal', 'value': 0}]
    assert df_formats[0]['conditional_formats'][1]['color'] == '#FFFFFF'
    assert df_formats[0]['conditional_formats'][1]['backgroundColor'] == '#000000'
    assert df_formats[0]['conditional_formats'][1]['invalidFilterColumnIDs'] == []

    # They should have different format_uuids
    assert df_formats[0]['conditional_formats'][0]['format_uuid'] != df_formats[0]['conditional_formats'][1]['format_uuid']


def test_create_backend_with_column_definitions_works_multiple_sheets():
    
    column_definitions= [
        [
            {
                'columns': ['A', 'B'],
                'conditional_formats': [{
                    'filters': [{'condition': 'greater', 'value': 5}], 
                    'font_color': '#c30010', 
                    'background_color': '#ffcbd1' 
                }] 
            }
        ], 
        [
            {
                'columns': ['A', 'B'],
                'conditional_formats': [{
                    'filters': [{'condition': 'greater_than_or_equal', 'value': 10}], 
                    'font_color': '#000000', 
                    'background_color': '#FFFFFF' 
                }] 
            },
        ]
    ]

    dfs = [df, df]
    mito_backend = get_mito_backend(*dfs, column_definitions=column_definitions)

    df_formats = mito_backend.steps_manager.curr_step.df_formats
    
    # Still generates valid, empty df_formats objects
    assert df_formats[0]['columns'] == {}
    assert df_formats[0]['headers'] == {}
    assert df_formats[0]['rows']['even'] == {}
    assert df_formats[0]['rows']['odd'] == {}
    assert df_formats[0]['border'] == {}

    # But now has conditional formats
    assert len(df_formats[0]['conditional_formats']) == 1
    assert df_formats[0]['conditional_formats'][0]['columnIDs'] == ['A', 'B']
    assert df_formats[0]['conditional_formats'][0]['filters'] == [{'condition': 'greater', 'value': 5}]
    assert df_formats[0]['conditional_formats'][0]['color'] == '#c30010'
    assert df_formats[0]['conditional_formats'][0]['backgroundColor'] == '#ffcbd1'
    assert df_formats[0]['conditional_formats'][0]['invalidFilterColumnIDs'] == []

    assert len(df_formats[1]['conditional_formats']) == 1
    assert df_formats[1]['conditional_formats'][0]['columnIDs'] == ['A', 'B']
    assert df_formats[1]['conditional_formats'][0]['filters'] == [{'condition': 'greater_than_or_equal', 'value': 10}]
    assert df_formats[1]['conditional_formats'][0]['color'] == '#000000'
    assert df_formats[1]['conditional_formats'][0]['backgroundColor'] == '#FFFFFF'
    assert df_formats[1]['conditional_formats'][0]['invalidFilterColumnIDs'] == []

    # They should have different format_uuids
    assert df_formats[0]['conditional_formats'][0]['format_uuid'] != df_formats[1]['conditional_formats'][0]['format_uuid']


INVALID_COLUMN_DEFINITIONS = [
    (
        [
            [
                {
                    'columns': ['A'],
                    'conditional_formats': [{
                        'filters': [{'condition': 'less_than_or_equal', 'value': 0}], 
                        'font_color': 'ABC', 
                        'background_color': '#000000' 
                    }] 
                },
            ]
        ], "set in column_definititon is not a valid hex color"
    ),
    (
        [
            [
                {
                    'columns': ['A'],
                    'conditional_formats': [{
                        'filters': [{'condition': 'less_than_or_equal', 'value': 0}],
                    }] 
                },
            ]
        ], 
        "column_definititon has invalid conditional_format rules. It must set the font_color, background_color, or both."
    ),
    (
        [
            [
                {
                    'columns': ['A'],
                    'conditional_formats': [{
                        'filters': [{'condition': 'less_than_or_equal', 'value': 0}],
                        'font_color': '#000000', 
                        'background_color': '#000000' 
                    }] 
                },
            ], 
            [
                {
                    'columns': ['A'],
                    'conditional_formats': [{
                        'filters': [{'condition': 'less_than_or_equal', 'value': 0}],
                        'font_color': '#000000', 
                        'background_color': '#000000' 
                    }] 
                },
            ], 
        ],
        "dataframes are provided."
    ),
    (
        [
            [
                {
                    'columns': ['D'],
                    'conditional_formats': [{
                        'filters': [{'condition': 'less_than_or_equal', 'value': 0}],
                        'font_color': '#000000', 
                        'background_color': '#000000' 
                    }] 
                },
            ], 
        ], 
        "don't exist in the dataframe."
    ),
    (
        [
            [
                {
                    'columns': ['A'],
                    'conditional_formats': [{
                        'filters': [{'condition': 'INVALID_FILTER', 'value': 0}],
                        'font_color': '#000000', 
                        'background_color': '#000000' 
                    }] 
                },
            ], 
        ], 
        "The condition INVALID_FILTER is not a valid filter condition."
    ),
]
@pytest.mark.parametrize("column_definitions,error", INVALID_COLUMN_DEFINITIONS)
def test_invalid_column_definitions(column_definitions, error):

    with pytest.raises(ValueError) as e_info:
        mito_backend = get_mito_backend(df, column_definitions=column_definitions)
        
    assert error in str(e_info)


def test_create_backend_with_column_definitions_does_not_error_with_mismatch_condition_and_column_type():
    column_definitions= [
        [
            {
                'columns': ['A'],
                'conditional_formats': [{
                    'filters': [{'condition': "string_does_not_contain", 'value': 0}],
                    'font_color': '#c30010', 
                    'background_color': '#ffcbd1'
                }] 
            },
        ], 
    ]

    mito_backend = get_mito_backend(df, column_definitions=column_definitions)
    df_formats = mito_backend.steps_manager.curr_step.df_formats
    assert len(df_formats[0]['conditional_formats']) == 1
    assert df_formats[0]['conditional_formats'][0]['columnIDs'] == ['A']
    assert df_formats[0]['conditional_formats'][0]['filters'] == [{'condition': "string_does_not_contain", 'value': 0}]
    assert df_formats[0]['conditional_formats'][0]['color'] == '#c30010'
    assert df_formats[0]['conditional_formats'][0]['backgroundColor'] == '#ffcbd1'
    assert df_formats[0]['conditional_formats'][0]['invalidFilterColumnIDs'] == []
        

def test_column_definitions_with_string_indexes_conditional_format_works():
    column_definitions= [
        [
            {
                'columns': ['A'],
                'conditional_formats': [{
                    'filters': [{'condition': 'number_exactly', 'value': 2}], 
                    'font_color': '#c30010', 
                    'background_color': '#ffcbd1' 
                }] 
            }
        ]
    ]


    mito_backend = get_mito_backend(pd.DataFrame({'A': [1, 2, 3]}, index=['S1', 'S2', 'S3']), column_definitions=column_definitions)
    sheet_json = mito_backend.get_shared_state_variables()['sheet_data_json']
    sheet_data = json.loads(sheet_json)[0]
    conditional_formatting_result = sheet_data['conditionalFormattingResult']
    assert len(conditional_formatting_result['results']) == 1
    assert len(conditional_formatting_result['results']['A']) == 1
    assert 'S2' in conditional_formatting_result['results']['A']







#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains utilities used in multiple sheet functions.
"""
from typing import Tuple, Union
import pandas as pd

def try_extend_series_to_index(series: pd.Series, index_to_fill: Union[pd.Index, pd.MultiIndex]) -> pd.Series:
    """
    Extends a given series to contain the entire index, filling the series
    with the first defined value.

    If the series has > 1 item in it, we assume it is already extended,
    and do not modify it.
    """
    if series.size > 1:
        return series
    return pd.Series([series.iloc[0]] * len(index_to_fill), index=index_to_fill)


def fill_series_with_one_index(series_list: Tuple[pd.Series, ...]) -> Tuple[pd.Series, ...]:
    """
    Extends all series in the series_list so that the indexes
    of every series is defined in each others. Uses the last
    defined index of the series to do so. Avoids the issue:

    pd.Series(data=[1]) + pd.Series(data=[1, 2]) = pd.Series(data=[1, Nan])

    """
    # Get all the one_index_series, so we can fill them
    one_index_series_list = [series for series in series_list if is_series_of_constant(series)]
    # Find a series to fill from (if we cannot, then we can just return)
    non_one_index_series_list = [series for series in series_list if not is_series_of_constant(series)]
    non_one_index_series = non_one_index_series_list[0] if len(non_one_index_series_list) > 0 else one_index_series_list[0]
    
    indexes_to_fill = non_one_index_series.index

    final_series_list = []
    for series in series_list:
        if len(series) == 1:
            final_series = try_extend_series_to_index(series, indexes_to_fill)
            final_series_list.append(final_series)
        else:
            final_series_list.append(series)
    
    return tuple(final_series_list)


def is_series_of_constant(series: pd.Series) -> bool:
    """
    A helper function for determining if a series has been constructed from
    a constant, which means that it is a series with just one element at index
    0.

    We need to make sure to extend these series, so that we can operate on
    them with sheet functions properly. 
    """
    return series.size == 1 and series.index.tolist() == [0]


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from typing import Any, Dict, List
from mitosheet.types import StepsManagerType


def get_all_params_for_step_type(params: Dict[str, Any], steps_manager: StepsManagerType) -> List[Any]:
    """
    Gets all the params for a given step type, in reverse order they were performed
    """
    step_type = params['step_type']

    # Loop over the steps backwards, so that we get the most recent ones first
    found_params: List[Any] = []
    for step in steps_manager.steps_including_skipped[:steps_manager.curr_step_idx + 1][::-1]:
        if step.step_type == step_type:
            found_params.append(step.params)

    # Return empty list if no params were found
    return found_params